# -*- coding: utf-8 -*-
"""
Helpers to initialize some things
"""
import logging

from optimus.logs import ColorizingStreamHandler
from optimus import utils

def init_logging(loglevel, printout=True, logfile=None):
    """
    Initialize the app's logger
    """
    rootlogger = logging.getLogger('optimus')
    rootlogger.setLevel(getattr(logging, loglevel))
    
    if not printout and not logfile:
        from StringIO import StringIO
        dummystream = StringIO()
        rootlogger.addHandler(logging.StreamHandler(dummystream))
    else:
        if printout:
            rootlogger.addHandler(ColorizingStreamHandler())
        if logfile:
            rootlogger.addHandler(logging.FileHandler(logfile))
    
    rootlogger.debug("Set logger level to: %s", loglevel)
    return rootlogger

def initialize(settings):
    """
    Init the needed directory structure
    """
    #utils.patch_webassets_bug(settings)
    utils.init_directory(settings.STATIC_DIR)
    utils.init_directory(settings.WEBASSETS_CACHE)

    if settings.FILES_TO_SYNC is not None:
        for item in settings.FILES_TO_SYNC:
            utils.synchronize_assets_sources(settings, *item)

def display_settings(settings, names):
    """
    Helper to display some settings if they are setted
    """
    logger = logging.getLogger('optimus')
    for item in names:
        logger.debug(" - Settings.%s = %s", item, getattr(settings, item, 'NOT SET'))



# -*- coding: utf-8 -*-
"""
Settings configuration management
"""
import os, imp, logging, sys

def import_project_module(name):
    """
    Load the given module name, actually only from the current directory (where the CLI 
    has been started)
    """
    project_directory = os.getcwd()
    
    logger = logging.getLogger('optimus')
    logger.info('Loading "%s" module', name)
    logger.info('Module searched in: %s', project_directory)
    
    fp, pathname, description = imp.find_module(name, [project_directory])
    try:
        settings_mod = imp.load_module(name, fp, pathname, description)
    except:
        logger.critical('Unable to load module file')
        raise
    finally:
        # Close fp explicitly.
        if fp:
            fp.close()

    return settings_mod



# -*- coding: utf-8 -*-
"""
Common bundles to use with webassets
"""
from webassets import Bundle

COMMON_BUNDLES = {
    'css_screen_common': Bundle(
        'css/screen.css',
        filters='yui_css',
        output='css/screen.min.css'
    ),
    'css_ie_common': Bundle(
        'css/ie.css',
        filters='yui_css',
        output='css/ie.min.css'
    ),
    'js_ie_common': Bundle(
        'js/modernizr.custom.js',
        'js/respond.src.js',
        filters='yui_js',
        output='js/ie.min.js'
    ),
    'js_jquery': Bundle(
        'js/jquery/jquery-1.7.1.js',
        filters='yui_js',
        output='js/jquery.min.js'
    ),
}



# -*- coding: utf-8 -*-
"""
Command line actions
"""
import datetime, logging, os, time

from pathtools.patterns import match_path

from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

from argh import arg

from optimus.builder.assets import build_assets
from optimus.builder.pages import PageBuilder
from optimus.conf import import_project_module
from optimus.init import init_logging, initialize, display_settings

class TemplatesWatchEventHandler(PatternMatchingEventHandler):
    """
    Template changes handler
    
    TODO: * Assets rebuild (in similar but separate and dedicated handler)
          * Find all tricks to have more granular builds (don't build everything for all changes)
    """
    def __init__(self, settings, logger, assets_env, pages_env, *args, **kwargs):
        self.settings = settings
        self.logger = logger
        self.assets_env = assets_env
        self.pages_env = pages_env
        super(TemplatesWatchEventHandler, self).__init__(*args, **kwargs)
    
    def on_moved(self, event):
        # We are only interested to the destination
        if match_path(event.dest_path, 
                included_patterns=self.patterns,
                excluded_patterns=self.ignore_patterns,
                case_sensitive=self.case_sensitive):
            self.logger.debug("Change detected from a move on: %s", event.dest_path)
            self.build_for_item(event.dest_path)
    
    def on_created(self, event):
        self.logger.debug("Change detected from a create on: %s", event.src_path)
        self.build_for_item(event.src_path)

    def on_modified(self, event):
        self.logger.debug("Change detected from an edit on: %s", event.src_path)
        self.build_for_item(event.src_path)

    def get_relative_path(self, path):
        if path.startswith(self.settings.TEMPLATES_DIR):
            return path[len(self.settings.TEMPLATES_DIR)+1:]
        elif path.startswith(self.settings.SOURCES_DIR):
            return path[len(self.settings.SOURCES_DIR)+1:]
        return path

    def build_for_item(self, path):
        """
        (Re)build all pages using the changed template
        
        ``path`` argument is a template path
        """
        rel_path = self.get_relative_path(path)
        
        # Search in the registry if the file is a knowed template dependancy
        if rel_path in self.pages_env.registry.elements:
            self.logger.debug("Build required for: %s", rel_path)
            
            requires = self.pages_env.registry.get_pages_from_dependency(rel_path)
            self.logger.debug("Requires for rebuild: %s", requires)
            self.pages_env.build_bulk(requires)
            
    
@arg('--settings', default='settings', help='Python path to the settings module')
@arg('--loglevel', default='info', choices=['debug','info','warning','error','critical'], help='The minimal verbosity level to limit logs output')
@arg('--logfile', default=None, help='A filepath that if setted, will be used to save logs output')
@arg('--silent', default=False, help="If setted, logs output won't be printed out")
def watch(args):
    """
    The watch action for the commandline
    """
    root_logger = init_logging(args.loglevel.upper(), printout=not(args.silent), logfile=args.logfile)
    settings = import_project_module(args.settings)
    display_settings(settings, ('DEBUG', 'PROJECT_DIR','SOURCES_DIR','TEMPLATES_DIR','PUBLISH_DIR','STATIC_DIR','STATIC_URL'))
    
    if hasattr(settings, 'PAGES_MAP'):
        root_logger.info('Loading external pages map')
        pages_map = import_project_module(settings.PAGES_MAP)
        setattr(settings, 'PAGES', pages_map.PAGES)
    
    assets_env = build_assets(settings)
    pages_env = PageBuilder(settings, assets_env=assets_env)
    pages_env.scan_bulk(settings.PAGES)
    
    observer = Observer()
    
    event_handler = TemplatesWatchEventHandler(settings, root_logger, assets_env, pages_env, patterns=['*.html'], ignore_patterns=None, ignore_directories=False, case_sensitive=False)
    
    observer.schedule(event_handler, settings.TEMPLATES_DIR, recursive=True)
    observer.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()



# -*- coding: utf-8 -*-
"""
Various helpers
"""
import logging, os, shutil

def init_directory(directory):
    logger = logging.getLogger('optimus')
    if not os.path.exists(directory):
        logger.debug('Creating directory: %s', directory)
        os.makedirs(directory)
        return True
    return False

def synchronize_assets_sources(settings, src, dest):
    """
    For now, this is just a rmtree/copytree of the given path
    
    TODO: In future, this should be a clean synchronize, like with rsync
    
    * ``src`` arg is allways a file path assumed to be located in the 
    ``settings.SOURCES_DIR``
    * ``dst`` is a file path that will be in 
    ``settings.STATIC_DIR``.
    """
    logger = logging.getLogger('optimus')
    source = os.path.join(settings.SOURCES_DIR, src)
    if not os.path.exists(source):
        logger.warning('The given source does not exist and so can not be synchronized : %s', source)
        return
    
    destination = os.path.join(settings.STATIC_DIR, src)
    if os.path.exists(destination):
        logger.debug('Removing old asset destination: %s', destination)
        shutil.rmtree(destination)
    logger.debug('Synchronizing asset from "%s" to "%s"', source, destination)
    shutil.copytree(source, destination)

def patch_webassets_bug(settings):
    """
    Clean some dirs to bypass a bug in the webassets dev version
    in debug mode
    
    This will not work with multiple page that not share all the same assets (the 
    previously generated will be deleted)
    """
    if settings.DEBUG == True:
        logger = logging.getLogger('optimus')
        logger.warning('Old webassets patch is running')
        if os.path.exists(settings.STATIC_DIR):
            shutil.rmtree(settings.STATIC_DIR)
        if os.path.exists(settings.WEBASSETS_CACHE):
            shutil.rmtree(settings.WEBASSETS_CACHE)



# -*- coding: utf-8 -*-
"""
A naive trick to manage multiple language types

Usage : ::

    class LangFr(LangBase):
        code = 'fr'
        alt_code = 'fr'
        label = 'France'

    class LangUsa(LangBase):
        code = 'usa'
        alt_code = 'en'
        label = 'Usa'

    class LangUk(LangBase):
        code = 'uk'
        alt_code = 'en'
        external_code = 'en'
        label = 'United Kingdom'

This trick should be replaced by Jinja+i18n
"""
class LangBase(object):
    label = None # Label to display
    code = None # Internal code
    alt_code = None # Alternative code, will be equal "code" if not setted
    external_code = None # External code for some external apps, will be equal "alt_code" if not setted
    
    def __init__(self):
        if self.alt_code is None:
            self.alt_code = self.code
        if self.external_code is None:
            self.external_code = self.code
    
    def __str__(self):
        return self.label.encode('utf-8')
    
    def __unicode__(self):
        return self.label



# -*- coding: utf-8 -*-
import copy, logging

from webassets import Environment as AssetsEnvironment

from optimus.builder.bundles import COMMON_BUNDLES

def build_assets(settings):
    """
    Initialize webassets environment and build assets
    """
    logger = logging.getLogger('optimus')
    if not settings.PAGES:
        logger.info("Asset management skipped as there are no enabled bundles")
        return None
    logger.info("Starting asset management")
    
    # Assets bundles
    AVAILABLE_BUNDLES = copy.deepcopy(COMMON_BUNDLES)
    AVAILABLE_BUNDLES.update(getattr(settings, 'EXTRA_BUNDLES', {}))
    
    # Initialize webassets environment
    assets_env = AssetsEnvironment()
    assets_env.debug = settings.DEBUG
    assets_env.url = settings.STATIC_URL
    assets_env.directory = settings.STATIC_DIR
    assets_env.load_path = [settings.SOURCES_DIR]
    assets_env.cache = settings.WEBASSETS_CACHE

    # Register enabled assets bundles
    for bundle_item in settings.ENABLED_BUNDLES:
        logger.debug("Registering bundle: %s", bundle_item)
        assets_env.register(bundle_item, AVAILABLE_BUNDLES[bundle_item])
        
    # for debugging purpopse
    for bundle_item in settings.ENABLED_BUNDLES:
        logger.info(" Processing: %s", assets_env[bundle_item].resolve_output())
        for url in assets_env[bundle_item].urls():
            logger.debug(" - %s", url)
    
    return assets_env
